/*
 * Stripe Snoop 1.5
 *
 * Check http://stripesnoop.sourceforge.net for more information
 * about the hardware this software interfaces with.
 *
 *
 * Stripe Snoop currently compiles on 3 Operating systems.
 * Select your target OS by uncommenting the correct #define
 *
 * Example: I want to compile for a Win98 box, so I uncomment
 *          the "#define WIN9X 1" line
 *
 * WIN9X - This compiles a 32 bit Windows Application, for
 * Windows 95, 98, 98se, and ME. It requires no extra DLLs.
 *
 * WINNT - This compiles a 32 bit Windows Application for
 * Windows NT 4.0, 2000, and XP (both Pro and Home). This code
 * is dependent on inpout32.dll.
 *
 * LINUX - Creates a binary for Linux (possibly other UNIX-Like
 * OSes, as long as #includes are adjusted accordingly). NOTE!
 * Since this program requires direct access to i/o ports, it
 * must be run as root, or use setuid root.
 *
 */

#define PORT 0x201

#define ODD 1
#define EVEN 0

/* select the OS to compile for? */
/*#define WIN9X 1*/
/*#define WINNT 1*/
#define LINUX

/* Standard Includes */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "parser.h"
#include "ss.h"

#ifndef LINUX
 /*if we aren't a Linux build...*/
 #include <conio.h>
#endif

/* necessary of a Win9X build*/
#ifdef WIN9X
 #define Inp32 _inp
#endif

/* necessary for a Win NT build*/
#ifdef WINNT
 /* prototypes for Inp32out32.dll calls */
 short _stdcall Inp32(short PortAddress);
 void _stdcall Out32(short PortAddress, short data);
#endif

#ifdef LINUX
 #include <sys/io.h>
 #include <unistd.h>
 #define Inp32 inb
#endif

int VERBOSE=0; /*verbose flag*/
int RAW=0; /*RAW Flag*/
int INPUT=0; /*console used to give bitstream*/
int FORCE=0;
int CHERRY=0;


int findSS( char* buffer) {
	int flag = 0;
	int k=0;
	for(k=0;k<240;k++) 	{
		if( (buffer[k] == 1) && (buffer[k + 1] == 1) && (buffer[k + 2] == 0) && (buffer[k + 3] == 1) && (buffer[k + 4] == 0))
			break;
	}
	if(k>=239) {
		flag=1;
	}

	if(flag == 1) {
		if (VERBOSE) printf("Start char not found\n");
		return -1;

	} else {
		if(VERBOSE) printf("Start char located at: %d\n", k);
		return k;
	}
}

int findES( char* buffer, int index) {
	int flag= 0;
	int k=0;
	for(k=index;k<=235;k+=5) 	{
		if ((buffer[k] == 1) && (buffer[k + 1] == 1) && (buffer[k + 2] == 1) && (buffer[k + 3] == 1) && (buffer[k + 4] == 1))
		break;
	}
	if(k>=239) 	{
		flag=1;
	}

	if(flag == 1) {
		if (VERBOSE) printf("End Char not found\n");
		return -1;
	} else {
		if(VERBOSE) printf("End Char located at: %d\n", k);
		return k;
	}
}

int lrcCheck(char *buffer, int start, int end) {
	char lrc[5];
	int i, j;
	for(i=0;i<5;i++)
		lrc[i]=0;

	/*Just XOR the Data Bits of each character, not the parity bit*/
	for(i=start;i<=end;i+=5) {

		for(j=0;j<4;j++)
    		lrc[j] = lrc[j] ^ buffer[i+j];
    }

	/*Calculate LRC Odd Parity bit*/
	i = 0;
	for(j = 0; j<4;j++)
	    if(lrc[j]==1) i++;

	if( (i % 2) == 0) lrc[4]=1;

	if(end+5 <=235) {
		/* valid LRC? */
		if ( (lrc[0] == buffer[end + 5])
			&& (lrc[1] == buffer[end + 6])
			&& (lrc[2] == buffer[end + 7])
			&& (lrc[3] == buffer[end + 8])
			&& (lrc[4] == buffer[end + 9]) )
			return 1;
	}
	return 0;
}

char * reverse(char *buffer) {
	char * temp;
	int i=0;
	temp = (char *) malloc(240);
	if(temp == NULL) {
		printf("Error: reverse(): Malloc failed");
		exit(1);
	}
	for(i=0;i<240;i++)
		temp[i]=buffer[239-i];
	return temp;
}

void goRaw()
{
	if(VERBOSE) printf("Raw Mode:\n");

	int e;
	int i=0;
	while(1) {
		do
		{
			e=Inp32(PORT);

		}while( (e & 32) !=0);

		if( (e & 16) ==0)
			printf("1");
		else
			printf("0");
		fflush(stdout);
			
		do
		{
			e=Inp32(PORT);
		}while( (e & 32) != 32);
	}
}

char * readInterface()
{
	char *cardRaw; /* keep this abstract*/
	int e, i;

	if( (cardRaw = (char *) malloc(240)) == NULL) {
		printf("Error: goParse(): Cannot malloc space for cardRaw\n");
		exit(1);
	}

	for(i=0;i<240;i++)
		cardRaw[i]=0;
	
	printf("Waiting for Card\n");

	for(i=0;i<240;i++) {
		/*trap the clock line*/
		do {
			e=Inp32(PORT);
		} while( (e & 32) !=0);
		/*store the value*/
		cardRaw[i]=e;
		do {
			e=Inp32(PORT);
		} while( (e & 32) != 32);
		/*done trapping the clock line*/
	}

	if (VERBOSE) printf("Creating Bitstream...\n");

	/* stripe it to a bitstream*/
	for(i = 0;i<240;i++) {
		cardRaw[i] = (cardRaw[i] & 16);
		if(cardRaw[i]==0) {
			cardRaw[i]=1;
		} else {
			cardRaw[i]=0;
		}
	}
	/* return the bitstream */
	return cardRaw;
}

char * goParseBCD(char * bitstream)
{
	int start, end;
	char * temp;	
		
	start = findSS(bitstream);
	end = findES(bitstream, start);
	/* is bitstream valid? */
	
	if(lrcCheck(bitstream, start, end) == 0 || start < 0 ||
	   end < 0 || (start >= end) ){
		if (VERBOSE) printf("LRC Check fails, Trying to parse backwards\n");
        	/* try backwards */
		temp = reverse(bitstream);
        	free(bitstream); /*Avoid memory leak*/
        	bitstream = temp;
		start = findSS(bitstream);
		end = findES(bitstream, start);
		if(lrcCheck(bitstream, start, end) == 0 || start < 0 || end < 0 || (start >= end) ) {
			printf("LRC check failed ion both directions\n");
			printf("I cannot read this card, no valid Track II\n");
			exit(1);
		}

	}
	return decodeBCD(bitstream, start, end);
}

char * goForceBCD(char * bitstream)
{

	int start=0;
	if(VERBOSE) printf("Forcing BCD Parsing...\n");
	if((start = findSS(bitstream)) == -1) {
		printf("No start character at all found in Bitstream. Nothing to force!\n");
		exit(1);
	}
	return decodeBCD(bitstream, start, 235);
}

bool parityCheck(char *bs, int type, int bpc)
{
	int i,j=0;
	for(i=0;i<bpc-1;i++)
		if(bs[i]==1) j++;
	if((j % 2) ==0 && bs[bpc-1]!=type)
		return false;
	return true;
}

char * decodeBCD(char *bitstream, int start,int end)
{
	char* cardOut;
	int i,j,k;
	char *ptrNext;
	char BCD[17];
	bool parOK;

	//strcpy(BCD, ".08.4..<2..:.6>.1..9.5=..3;.7..?");
	strcpy(BCD, "084<2:6>195=3;7?");
	if( (cardOut= (char *) malloc(41)) == NULL) {
		printf("Error: decodeBCD(): Cannot malloc space for cardOut\n");
		exit(1);
	}
	memset(cardOut,0,41);
	ptrNext = cardOut;

	for(i=start;i<=end;i+=5) {
		k=0;
		for(j=0;j<4;j++)
			k=k*2+bitstream[i+j];
		parOK=parityCheck(&bitstream[i], ODD, 5);
		if(!parOK) {
			if(VERBOSE || !FORCE)
				printf("Parity Check Failed at %d\n",i);
			if(!FORCE) exit(1);
		}
		if(k>=0 && k<=15)
			*ptrNext++=BCD[k];
		else
			*ptrNext++='X';
	}

	return(cardOut);
}

int isBCD(char * s)
{
	int  i;
	for(i=0;i<strlen(s);i++)
		if(s[i]<48 || s[i] >63) {
			if(VERBOSE)
				printf("Invalid BCD character \'%c\' detected at %d\n",s[i],i);
			return -1;
		}
	return 1;
}

char * handleCherry(char * cardBoth)
{
	//ok, are we even dealing with a multi track card?
	char temp[500];
	memset(temp,0,500);
	int i,j;

	if(strlen(cardBoth)>40) {
		//We are going to attempt to find Track 2
		if(VERBOSE) printf("\nMultiple Tracks Detected! Stripe Snoop only supports banking cards in this mode.\n");
		//attempt to locate field seperator of Track2
		for(i=strlen(cardBoth)-1;i>=0;i--)
			if(cardBoth[i]=='=') break;
		if(i!=0) {
			//look for an account number before the '='
			for(j=13;j<=16;j++) {
				strncpy(temp,&cardBoth[i-j],j);
    				if(mod10check(temp)==0) {
					memset(temp,0,500);
					temp[0]=';';
					strcpy(&temp[1],&cardBoth[i-j]);
					temp[strlen(temp)]='?';
					strcpy(cardBoth,temp);
					cardBoth[strlen(temp)]=0;
					return cardBoth;
				}
			}
		}

	} else {
		//are we dealing with a valid Track 2 only?
		//ie only BCD
		if(isBCD(cardBoth)==1) {
			//ok we have BCD track 2. Just stick the ";" and "?" characters on
			temp[0]=';';
			strcpy(&temp[1],cardBoth);
			temp[strlen(cardBoth)]='?';
			strcpy(cardBoth,temp);
			return cardBoth;
		} else {
			printf("Some characters are not BCD\n");
		}
	}
	return NULL;
}
/*----------------------------------------------------------------------MAIN*/
int main(int argc, char* argv[])
{

	int i,j,c;
	char *bitstream; /* bitstream of the card*/
	char *cardOut;

	cardType result;
	track2 t2;
	
	//parse commandline options!

	for(c=2;c<=argc;c++)
	{
		if(strcmp("-r",argv[c-1])==0)
			RAW = 1;
		if(strcmp("-v",argv[c-1])==0)
			VERBOSE = 1;
		if(strcmp("-i",argv[c-1])==0)
			INPUT = 1;		
		if(strcmp("-F",argv[c-1])==0) {
			FORCE=1;
		}
		if(strcmp("-c",argv[c-1])==0) {
			CHERRY=1;
		}
	}
	if(!RAW) {
		printf("Stripe Snoop Version 1.5\n");
		printf("http://stripesnoop.sourceforge.net  Acidus@yak.net\n\n");
	}
	#ifdef LINUX
		if(!INPUT && !CHERRY) {
			// Notify Linux what ports we will need
			if(ioperm(PORT,1,1)) {
				printf("Error: main(): ioperm error.\nStripe Snoop must be run root\n");
				exit(1);
			}
		}
	#endif

//---------------------------------------STAGE 1
//---------------------------------------Capture or produce Bitstream
		
	if(RAW == 1) {
		goRaw();
	} else if (INPUT == 1){
		if(VERBOSE) printf("INPUT MODE: Waiting for bitstream on stdin\n");
		/* we are getting our bitstream from the console*/
		if( (bitstream = (char *) malloc(240)) == NULL) {
			printf("Error: main(): Cannot malloc space for bitstream\n");
			exit(1);
		}
		memset(bitstream,0,240);
		fgets(bitstream,240,stdin);
		/*map to 1 and 0 */
		for(c=0;c<240;c++)
			if(c>0) bitstream[c]=bitstream[c]- 48;

	} else if(CHERRY==1) {
		//we will use bitstream to hold the contents of the card!
		//79 characters  +40
		printf("Please Note support for keyboard based readers is very primitive\n");
		printf("Swipe card in reader and press enter\n");
		
		if( (bitstream = (char *) malloc(120)) == NULL) {
			printf("Error: main(): Cannot malloc space for bitstream\n");
			exit(1);
		}
		memset(bitstream,0,120);
		fgets(bitstream,120,stdin);
		bitstream[strlen(bitstream)-1]='\0'; //remove trailing CR
		cardOut= handleCherry(bitstream);
	} else {
		/* get bitstream from Interface */
		bitstream = readInterface();

	}
//---------------------------------------STAGE 2
//---------------------------------------Parse the bitstream
	
	if(!CHERRY)
		if(!FORCE) {
			cardOut = goParseBCD(bitstream);
		} else {
			cardOut = goForceBCD(bitstream);
		}
//---------------------------------------STAGE 3
//---------------------------------------Analyze the Bitstream

	if(cardOut != NULL) {
		printf("Card Contents: \"%s\"\n\n",cardOut);
	
		//extract the fields from track 2
		t2 = extractTrack2(cardOut);
		//Look it up in the database!
		result = lookup2(t2);
		if(result.numFields == 0)
			printf("No Match in Database\n");
		else {
			printf("Possibly a \"%s\"\n",result.name);

			//find the longest name;
			c=0;
			for(i=0; i<result.numFields; i++)
				if(strlen(result.getName(i))>c)
					c=strlen(result.getName(i));
			c = ((int) c+1) / 8;


			for(i=0; i<result.numFields; i++) {

				printf("%s:",result.getName(i));
				for(j = (int) (strlen(result.getName(i))+1) /8; j<=c;j++)

					printf("\t");
				printf("%s\n",result.getData(i));
			}
			if(result.notes != NULL && strlen(result.notes) >0)
				printf("\nCard Notes: %s\n",result.notes);

			if(result.unknowns) {
				printf("\nUNKNOWNS FOUND - This card contains data that has not yet been decoded.\n");
				printf("Specifically: %s\n",result.untext);
				printf("\nPlease contact the developers so they can improve the program\n");
			}
		}
	}
	return 0;
}
